@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2024. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -90,8 +90,8 @@ The `init` module interprets the following command-line flags:
   It defaults to `strict` from OTP 27 and this option is scheduled for removal
   in OTP 28.
 
-- **`-epmd_module Module`** - Specifies the module to use for registration and
-  lookup of node names. Defaults to `erl_epmd`.
+- **`-epmd_module Module`** - This flag is deprecated and has been replaced by
+  the `kernel` application parameter [`epmd_module`](`e:kernel:kernel_app.md#epmd_module`).
 
 - **`-eval Expr`** - Scans, parses, and evaluates an arbitrary expression `Expr`
   during system initialization. If any of these steps fail (syntax error, parse
@@ -697,6 +697,11 @@ boot_loop(BootPid, State) ->
 	    notify(State#state.subscribed),
 	    loop(State#state{status = {started,PS},
 			     subscribed = []});
+        {'EXIT',BootPid,terminating} ->
+            %% If BootPid exited with terminating, then AC is about to
+            %% terminate and thus the entire system. We ignore this here
+            %% in order to give kernel a chance to flush logs to disk.
+            boot_loop(BootPid, State);
 	{'EXIT',BootPid,Reason} ->
 	    % erlang:display({"init terminating in do_boot",Reason}),
 	    crash("Runtime terminating during boot", [Reason]);
